// Listing 3.2: Arbeiten mit dem Schlsselbund: Suchen, Aktualisieren, Eintragen (aus [Appl_13.12])

#import <UIKit/UIKit.h>
#import <Security/Security.h>
 
// Definiere eine Objective-C Wrapper-Klasse fr den Code des Keychain Services:
@interface KeychainWrapper : NSObject {
    NSMutableDictionary *keychainData;
    NSMutableDictionary *genericPasswordQuery;
}
 
@property (nonatomic, strong) NSMutableDictionary *keychainData;
@property (nonatomic, strong) NSMutableDictionary *genericPasswordQuery;
 
- (void)mySetObject:(id)inObject forKey:(id)key;
- (id)myObjectForKey:(id)key;
- (void)resetKeychainItem;
 
@end

// ==============================
 
// Eindeutiger Identifier fr den Schlsselbund-Eintrag:
static const UInt8 kKeychainItemIdentifier[] = "example.entwickler.demo.KeychainUI\0";
 
@interface KeychainWrapper (PrivateMethods)

// Zwei Methoden zur Umwandlung der Dictionaries zwischen den Formaten des
// View-Controllers (NSString *) und des Keychain Services API:
- (NSMutableDictionary *)secItemFormatToDictionary:(NSDictionary *)dictionaryToConvert;
- (NSMutableDictionary *)dictionaryToSecItemFormat:(NSDictionary *)dictionaryToConvert;

// Methode zum Schreiben der Daten in den Schlsselbund:
- (void)writeToKeychain;
 
@end
 
// ==============================

@implementation KeychainWrapper
 
// Synthetisieren von Getter und Setter:
@synthesize keychainData, genericPasswordQuery;
 
- (id)init
{
    if ((self = [super init])) {
 
        OSStatus keychainErr = noErr;
        
        // Aufbau des Keychain Search Dictionary:
        genericPasswordQuery = [[NSMutableDictionary alloc] init];
        
        // Der Eintrag ist ein generisches Passwort:
        [genericPasswordQuery setObject:(id)kSecClassGenericPassword
                              forKey:(id)kSecClass];
                                 
        // Im kSecAttrGeneric-Attribut wird ein String gespeichert, der 
        // als eindeutiger Identifier fr den Eintrag dient.
        // Der String wird zuerst in ein NSData-Objeckt umgewandelt:
        NSData *keychainItemID = [NSData dataWithBytes:kKeychainItemIdentifier
                                  length:strlen((const char *)kKeychainItemIdentifier)];
        [genericPasswordQuery setObject:keychainItemID 
                              forKey:(id)kSecAttrGeneric];
        
        // Es werde nur die Attribute der ersten Fundstelle bentigt:
        [genericPasswordQuery setObject:(id)kSecMatchLimitOne
                              forKey:(id)kSecMatchLimit];
        
        // Rckgabe der Attribute des Eintrags, das Passwort wird von der 
        // Methode secItemFormatToDictionary: geholt:
        [genericPasswordQuery setObject:(id)kCFBooleanTrue
                              forKey:(id)kSecReturnAttributes];
 
        // Initialisierung des Dictionarys fr die Rckgabewerte:
        NSMutableDictionary *outDictionary = nil;
        
        // Wenn der Eintrag existiert werden seine Attribute zurckgegeben:
        keychainErr = SecItemCopyMatching((CFDictionaryRef)genericPasswordQuery,
                                           (CFTypeRef *)&outDictionary);
                                         
        if (keychainErr == noErr) {
            // Alles in Ordnung, das Daten-Dictionary wird in das vom 
            // View-Controller verwendete Format umgewandelt:
            self.keychainData = [self secItemFormatToDictionary:outDictionary];
        } else if (keychainErr == errSecItemNotFound) {
            // Der Eintrag wurde nicht gefunden, es werden die 
            // Default-Werte eingetragen:
            [self resetKeychainItem];
        } else {
            // Jede andere Fehler darf nicht vorkommen:
            NSAssert(NO, @"Unerwarteter Fehler.\n");
        }
    }
    return self;
}
 
// Implementierung der Methode mySetObject:forKey um die Attribute in den 
// Schlsselbund zu schreiben:
- (void)mySetObject:(id)inObject forKey:(id)key
{
    if (inObject == nil) return;
    id currentObject = [keychainData objectForKey:key];
    if (![currentObject isEqual:inObject])
    {
        [keychainData setObject:inObject forKey:key];
        [self writeToKeychain];
    }
}
 
// Implementierung der Methode myObjectForKey: zum Lesen eines 
// Attribut-Werts aus dem Dictionary:
- (id)myObjectForKey:(id)key
{
    return [keychainData objectForKey:key];
}
 
// Reset der Werte im Schlsselbund-Eintrag oder Anlegen eines neuen Eintrags, 
// falls noch keiner vorhanden ist
 
- (void)resetKeychainItem
{
    if (!keychainData) {
        // Anlegen des keychainData-Dictionaries, weil noch keins existiert:
        self.keychainData = [[NSMutableDictionary alloc] init];
    }
    else if (keychainData) {
        // Umwandeln der Daten des keychainData-Dictionaries in das Format 
        // fr die Abfrage und Speicherung des Ergebnisses in tmpDictionary:
        NSMutableDictionary *tmpDictionary =
                            [self dictionaryToSecItemFormat:keychainData];
        // Lschen des Eintrags vor dem Reset der Werte
        NSAssert(SecItemDelete((CFDictionaryRef)tmpDictionary) == noErr,
                               @"Fehler beim Lschen des aktuellen Eintrags." );
    }
 
    // Default-Daten fr den Schlsselbund-Eintrag:
    [keychainData setObject:@"Item Label" forKey:(id)kSecAttrLabel];
    [keychainData setObject:@"Item Beschreibung" forKey:(id)kSecAttrDescription];
    [keychainData setObject:@"Account" forKey:(id)kSecAttrAccount];
    [keychainData setObject:@"Service" forKey:(id)kSecAttrService];
    [keychainData setObject:@"Kommentar" forKey:(id)kSecAttrComment];
    [keychainData setObject:@"Password" forKey:(id)kSecValueData];
}
 
// Implementierung der Methode dictionaryToSecItemFormat: um die zum
// Schlsselbund-Eintrag hinzuzufgenden Attribute in das vom Keychain Service
// bentigte Dictionary-Format umzuwandeln:
- (NSMutableDictionary *)dictionaryToSecItemFormat:(NSDictionary *)dictionaryToConvert
{
    // Die Methode muss mit einem korret gefllten Dictionary aufgerufen 
    // werden, dass alle bentigten Schlssel-Wert-Paare fr eine 
    // Schlsselbund-Suche enthlt.
 
    // Anlegen des Dictionaries fr die Rckgabewerte:
    NSMutableDictionary *returnDictionary = [NSMutableDictionary 
                                             dictionaryWithDictionary:dictionaryToConvert];
 
    // Hinzufgen der Klasse und generischen Attribute des 
    // Keychain-Eintrags:
    NSData *keychainItemID = [NSData dataWithBytes:kKeychainItemIdentifier
                              length:strlen((const char *)kKeychainItemIdentifier)];
    [returnDictionary setObject:keychainItemID forKey:(id)kSecAttrGeneric];
    [returnDictionary setObject:(id)kSecClassGenericPassword forKey:(id)kSecClass];
 
    // Umwandeln des Passworts von NSString in NSData und die 
    // API-Anforderungen zu erfllen:
    NSString *passwordString = [dictionaryToConvert objectForKey:(id)kSecValueData];
    [returnDictionary setObject:[passwordString dataUsingEncoding:NSUTF8StringEncoding]
                                 forKey:(id)kSecValueData];
    return returnDictionary;
}
 
// Implementierung der Methode secItemFormatToDictionary:, die das 
// Dictionary mit den Attributen aus dem Schlsselbund-Eintrag nimmt, das 
// Passwort aus dem Schlsselbund holt und zum Attribut-Dictionary hinzufgt:

- (NSMutableDictionary *)secItemFormatToDictionary:(NSDictionary *)dictionaryToConvert
{
    // Die Methode muss mit einem korret gefllten Dictionary aufgerufen 
    // werden, dass alle bentigten Schlssel-Wert-Paare fr eine 
    // Schlsselbund-Suche enthlt.
 
    // Anlegen des Dictionaries fr die Rckgabewerte:
    NSMutableDictionary *returnDictionary = [NSMutableDictionary
                                             dictionaryWithDictionary:dictionaryToConvert];
 
    // Bevor das Passwort geholt werden kann, mssen der Suchschlssel und 
    // die Klassenattribute zur Abfrage des Passworts eingetragen werden:
    [returnDictionary setObject:(id)kCFBooleanTrue forKey:(id)kSecReturnData];
    [returnDictionary setObject:(id)kSecClassGenericPassword forKey:(id)kSecClass];
    
    // Aufruf der Keychain Services zum Holen des Passworts:
    NSData *passwordData = NULL;
    OSStatus keychainError = noErr; //
    keychainError = SecItemCopyMatching((CFDictionaryRef)returnDictionary,
                                         (CFTypeRef *)&passwordData);
    if (keychainError == noErr) {
        // Alles in Ordnung, der nicht mehr bentigte kSecReturnData-Key 
        // kann gelscht werden:
        [returnDictionary removeObjectForKey:(id)kSecReturnData];
 
        // Umwandeln des Passworts in einen NSString und Eintragen in das 
        // Dictionary fr die Rckgabe:
        NSString *password = [[[NSString alloc] initWithBytes:[passwordData bytes]
                              length:[passwordData length] encoding:NSUTF8StringEncoding] autorelease];
        [returnDictionary setObject:password forKey:(id)kSecValueData];
    }
    else if (keychainError == errSecItemNotFound) {
        // Wenn nichts gefunden wurde, ist nichts zu tun
        NSAssert(NO, @"Nichts in der Keychain gefunden.\n");
    }
    else {
        // Jede andere Fehler darf nicht vorkommen:
        NSAssert(NO, @"Unerwarteter Fehler.\n");
    }
 
    return returnDictionary;
}

// Implementierung der Methode writeToKeychain, die von mySet-Object aufgerufen 
// wird, die wiederum vom UI aufgerufen wird, wenn neue Daten in den Schlsselbund
// eingetragen werden mssen.
// Die Methode aktualisiert einen vorhandenen Schlsselbund-Eintrag oder legt einen
// neuen Eintrag an, wenn noch keiner vorhanden ist.
// Beim neu anlegen werden die bergebenen Attribute eingetragen und alle anderen
// Attribute auf Default-Werte gesetzt.
 - (void)writeToKeychain
{
    NSDictionary *attributes = nil;
    NSMutableDictionary *updateItem = nil;
 
    if (SecItemCopyMatching((CFDictionaryRef)genericPasswordQuery,
                             (CFTypeRef *)&attributes) == noErr) {
        // Der Eintrag existiert bereits und muss aktualisiert werden:
            
        // Die Attribute aus dem Schlsselbund werden in das Dictionary fr 
        // die Aktualisierung eingetragen:
        updateItem = [NSMutableDictionary dictionaryWithDictionary:attributes];
 
        // Die Klassen-Werte aus dem generischen Password-Query-Dictionary werden
        // in das Dictionary fr die Aktualisierung eingetragen
        [updateItem setObject:[genericPasswordQuery objectForKey:(id)kSecClass]
                                                    forKey:(id)kSecClass];
 
        // Initialisieren des Dictionayries mit den neuen Werten fr die Attribute:
        NSMutableDictionary *tempCheck = [self dictionaryToSecItemFormat:keychainData];
        
        // Lschen der Klasse, die kein Keychain-Attribut ist:
        [tempCheck removeObjectForKey:(id)kSecClass];
 
        // Es kann jeweils nur ein einziger Eintrag aktualisiert werden
        NSAssert(SecItemUpdate((CFDictionaryRef)updateItem,
                (CFDictionaryRef)tempCheck) == noErr,
                @"Aktualisierung des Schlsselbunds fehlgeschlagen." );
    }
    else {
        // Es gibt keine passenden Eintrag, also muss ein neuer erstellt werden.
        // Die neuen Werte wurden in mySetObject zum  keychainData-Dictionary hinzugefgt,
        // die restlichen Default-Werte vor dem Aufruf hinzugefgt.
        
        // Da kein Pointer fr den neuen Eintrag bentigt wird, wird als zweiter Parameter
        // NULL bergeben.
         NSAssert(SecItemAdd((CFDictionaryRef)[self dictionaryToSecItemFormat:keychainData],
                             NULL) == noErr, @"Hinzufgen zum Schlsselbund fehlgeschlagen." );
    }
}
 
 
@end
